@@ -8,6 +8,7 @@ import (
 	"io"            // io.WriteString()
 	"io/ioutil"     // ioutil.ReadFile()
 	"math/rand"     // rand.Seed(), ...
+	"net"           // net.Conn
 	"net/http"      // http server
 	"os"            // os.Exit(), os.Signal, os.Stderr, ...
 	"os/exec"       // os.Exec()
@@ -36,18 +37,17 @@ import (
 // Types
 type arrayFlags []string
 
-type resync struct {
-	now int64
-	min int64
+type sockAccepted struct {
+	conn net.Conn
+	err  error
 }
 
 type tunedState struct {
 	// label2value map
 	nodeLabels map[string]string
 	// namespace/podname -> label2value map
-	podLabels         map[string]map[string]string
-	podLabelsPullTime int64
-	change            struct {
+	podLabels map[string]map[string]string
+	change    struct {
 		// did node labels change?
 		node bool
 		// did pod labels change?
@@ -59,25 +59,29 @@ type tunedState struct {
 
 // Constants
 const (
-	resyncPeriodNodeDefault = 60
-	resyncPeriodNodeMax     = 3600     // maximum resync node period due to exponential backoff
-	resyncPeriodPodDefault  = 3600 * 8 // rely on watches and only pull pod labels every resyncPeriodPodDefault [s]
+	resyncPeriodDefault = 60
+	sleepRetry          = 5
 	// Minimum interval between writing changed node/pod labels for tuned daemon in [s]
 	labelDumpInterval      = 5
-	PNAME                  = "openshift-tuned"
+	programName            = "openshift-tuned"
 	tunedActiveProfileFile = "/etc/tuned/active_profile"
 	tunedProfilesConfigMap = "/var/lib/tuned/profiles-data/tuned-profiles.yaml"
 	tunedProfilesDir       = "/etc/tuned"
+	openshiftTunedRunDir   = "/run/" + programName
+	openshiftTunedPidFile  = openshiftTunedRunDir + "/" + programName + ".pid"
+	openshiftTunedSocket   = "/var/lib/tuned/openshift-tuned.sock"
 )
 
 // Global variables
 var (
 	done               = make(chan bool, 1)
+	tunedExit          = make(chan bool, 1)
 	terminationSignals = []os.Signal{syscall.SIGHUP, syscall.SIGINT, syscall.SIGTERM, syscall.SIGQUIT}
 	fileNodeLabels     = "/var/lib/tuned/ocp-node-labels.cfg"
 	filePodLabels      = "/var/lib/tuned/ocp-pod-labels.cfg"
 	fileWatch          arrayFlags
-	version            string // PNAME version
+	version            string // programName version
+	cmd                *exec.Cmd
 	// Flags
 	boolVersion = flag.Bool("version", false, "show program version and exit")
 )
@@ -105,8 +109,8 @@ func (a *arrayFlags) Set(value string) error {
 
 func parseCmdOpts() {
 	flag.Usage = func() {
-		fmt.Fprintf(os.Stderr, "Usage: %s [options] <NODE>\n", PNAME)
-		fmt.Fprintf(os.Stderr, "Example: %s b1.lan\n\n", PNAME)
+		fmt.Fprintf(os.Stderr, "Usage: %s [options] <NODE>\n", programName)
+		fmt.Fprintf(os.Stderr, "Example: %s b1.lan\n\n", programName)
 		fmt.Fprintf(os.Stderr, "Options:\n")
 
 		flag.PrintDefaults()
@@ -129,6 +133,17 @@ func signalHandler() chan os.Signal {
 	return sigs
 }
 
+func newUnixListener(addr string) (net.Listener, error) {
+	if err := os.Remove(addr); err != nil && !os.IsNotExist(err) {
+		return nil, err
+	}
+	l, err := net.Listen("unix", addr)
+	if err != nil {
+		return nil, err
+	}
+	return l, nil
+}
+
 func logsCoexist() {
 	flag.Set("logtostderr", "true")
 	flag.Parse()
@@ -179,10 +194,6 @@ func getConfig() (*rest.Config, error) {
 	return nil, fmt.Errorf("Could not locate a kubeconfig")
 }
 
-func getJitter(period int64, factor float64) int64 {
-	return rand.Int63n(int64(float64(period)*factor+1)) - int64(float64(period)*factor/2)
-}
-
 func profilesExtract() error {
 	glog.Infof("Extracting tuned profiles")
 
@@ -202,8 +213,7 @@ func profilesExtract() error {
 		profileDir := fmt.Sprintf("%s/%s", tunedProfilesDir, key)
 		profileFile := fmt.Sprintf("%s/%s", profileDir, "tuned.conf")
 
-		err = mkdir(profileDir)
-		if err != nil {
+		if err = mkdir(profileDir); err != nil {
 			return fmt.Errorf("Failed to create tuned profile directory '%s': %v", profileDir, err)
 		}
 
@@ -212,24 +222,118 @@ func profilesExtract() error {
 			return fmt.Errorf("Failed to create tuned profile file '%s': %v", profileFile, err)
 		}
 		defer f.Close()
-		_, err = f.WriteString(value)
-		if err != nil {
+		if _, err = f.WriteString(value); err != nil {
 			return fmt.Errorf("Failed to write tuned profile file '%s': %v", profileFile, err)
 		}
 	}
 	return nil
 }
 
+func openshiftTunedPidFileWrite() error {
+	if err := mkdir(openshiftTunedRunDir); err != nil {
+		return fmt.Errorf("Failed to create %s run directory '%s': %v", programName, openshiftTunedRunDir, err)
+	}
+	f, err := os.Create(openshiftTunedPidFile)
+	if err != nil {
+		return fmt.Errorf("Failed to create openshift-tuned pid file '%s': %v", openshiftTunedPidFile, err)
+	}
+	defer f.Close()
+	if _, err = f.WriteString(strconv.Itoa(os.Getpid())); err != nil {
+		return fmt.Errorf("Failed to write openshift-tuned pid file '%s': %v", openshiftTunedPidFile, err)
+	}
+	return nil
+}
+
+func tunedCreateCmd() *exec.Cmd {
+	return exec.Command("/usr/sbin/tuned", "--no-dbus")
+}
+
+func tunedRun() {
+	glog.Infof("Starting tuned...")
+
+	defer func() {
+		tunedExit <- true
+	}()
+
+	cmdReader, err := cmd.StderrPipe()
+	if err != nil {
+		glog.Errorf("Error creating StderrPipe for tuned: %v", err)
+		return
+	}
+
+	scanner := bufio.NewScanner(cmdReader)
+	go func() {
+		for scanner.Scan() {
+			fmt.Printf("%s\n", scanner.Text())
+		}
+	}()
+
+	err = cmd.Start()
+	if err != nil {
+		glog.Errorf("Error starting tuned: %v", err)
+		return
+	}
+
+	err = cmd.Wait()
+	if err != nil {
+		// The command exited with non 0 exit status, e.g. terminated by a signal
+		glog.Errorf("Error waiting for tuned: %v", err)
+		return
+	}
+
+	return
+}
+
+func tunedStop(s *sockAccepted) error {
+	if cmd == nil {
+		// Looks like there has been a termination signal prior to starting tuned
+		return nil
+	}
+	if cmd.Process != nil {
+		glog.V(1).Infof("Sending TERM to PID %d", cmd.Process.Pid)
+		cmd.Process.Signal(syscall.SIGTERM)
+	} else {
+		// This should never happen
+		return fmt.Errorf("Cannot find the tuned process!")
+	}
+	// Wait for tuned process to stop -- this will enable node-level tuning rollback
+	<-tunedExit
+	glog.V(1).Infof("Tuned process terminated")
+
+	if s != nil {
+		// This was a socket-initiated shutdown; indicate a successful settings rollback
+		ok := []byte{'o', 'k'}
+		_, err := (*s).conn.Write(ok)
+		if err != nil {
+			return fmt.Errorf("Cannot write a response via %q: %v", openshiftTunedSocket, err)
+		}
+	}
+
+	return nil
+}
+
 func tunedReload() error {
-	var stdout, stderr bytes.Buffer
+	if cmd == nil {
+		// Tuned hasn't been started by openshift-tuned, start it
+		cmd = tunedCreateCmd()
+		go tunedRun()
+		return nil
+	}
 
 	glog.Infof("Reloading tuned...")
-	cmd := exec.Command("/usr/sbin/tuned", "--no-dbus")
-	cmd.Stdout = &stdout
-	cmd.Stderr = &stderr
-	err := cmd.Run()
-	fmt.Fprintf(os.Stderr, "%s", stderr.String()) // do not use glog.Infof(), tuned has its own timestamping
-	return err
+
+	if cmd.Process != nil {
+		glog.Infof("Sending HUP to PID %d", cmd.Process.Pid)
+		err := cmd.Process.Signal(syscall.SIGHUP)
+		if err != nil {
+			return fmt.Errorf("Error sending SIGHUP to PID %d: %v\n", cmd.Process.Pid, err)
+		}
+	} else {
+		// This should never happen
+		return fmt.Errorf("Cannot find the tuned process!")
+	}
+
+	return nil
 }
 
 func nodeLabelsGet(clientset *kubernetes.Clientset, nodeName string) (map[string]string, error) {
@@ -539,7 +643,6 @@ func timedTunedReloader(tuned *tunedState) (err error) {
 
 func pullLabels(clientset *kubernetes.Clientset, tuned *tunedState, nodeName string) error {
 	// Resync period elapsed, force-pull node and pod labels
-	glog.V(2).Infof("Pulling node labels")
 	nodeLabels, err := nodeLabelsGet(clientset, nodeName)
 	if err != nil {
 		return err
@@ -549,55 +652,40 @@ func pullLabels(clientset *kubernetes.Clientset, tuned *tunedState, nodeName str
 		tuned.change.node = true
 	}
 
-	nowUnix := time.Now().Unix()
-	if nowUnix >= tuned.podLabelsPullTime {
-		glog.V(2).Infof("Pulling pod labels")
-		// Pull for pod labels was done >= resyncPeriodPodDefault
-		podLabels, err := podLabelsGet(clientset, nodeName)
-		if err != nil {
-			return err
-		}
-		tuned.setNextPodLabelsPullTime(nowUnix)
-		if !reflect.DeepEqual(podLabels, tuned.podLabels) {
-			tuned.podLabels = podLabels
-			tuned.change.pod = true
-		}
+	podLabels, err := podLabelsGet(clientset, nodeName)
+	if err != nil {
+		return err
+	}
+	if !reflect.DeepEqual(podLabels, tuned.podLabels) {
+		tuned.podLabels = podLabels
+		tuned.change.pod = true
 	}
-
 	return nil
 }
 
 func pullResyncPeriod() int64 {
 	var (
 		err                  error
-		resyncPeriodDuration int64 = resyncPeriodNodeDefault
+		resyncPeriodDuration int64 = resyncPeriodDefault
 	)
 	if os.Getenv("RESYNC_PERIOD") != "" {
 		resyncPeriodDuration, err = strconv.ParseInt(os.Getenv("RESYNC_PERIOD"), 10, 64)
 		if err != nil {
-			glog.Errorf("Error: cannot parse RESYNC_PERIOD (%s), using %d", os.Getenv("RESYNC_PERIOD"), resyncPeriodNodeDefault)
-			resyncPeriodDuration = resyncPeriodNodeDefault
+			glog.Errorf("Error: cannot parse RESYNC_PERIOD (%s), using %d", os.Getenv("RESYNC_PERIOD"), resyncPeriodDefault)
+			resyncPeriodDuration = resyncPeriodDefault
 		}
 	}
 
+	// Add some randomness to the resync period, so that we don't end up in a lockstep querying the API server
+	resyncPeriodDuration += rand.Int63n(resyncPeriodDuration/5+1) - resyncPeriodDuration/10
 	return resyncPeriodDuration
 }
 
-func pullResyncPeriodWithJitter() int64 {
-	resyncPeriodDuration := pullResyncPeriod()
-
-	// Add some randomness to the resync period, so that we don't end up in lockstep querying the API server
-	return resyncPeriodDuration + getJitter(resyncPeriodDuration, 0.3)
-}
-
-func (tuned *tunedState) setNextPodLabelsPullTime(secsSinceEpoch int64) {
-	tuned.podLabelsPullTime = secsSinceEpoch + resyncPeriodPodDefault + getJitter(resyncPeriodPodDefault, 0.3) // try to avoid lockstep
-}
-
-func (resyncPeriod *resync) changeWatcher() (err error) {
+func changeWatcher() (err error) {
 	var (
 		tuned tunedState
 		wPod  watch.Interface
+		lStop bool
 	)
 
 	nodeName := flag.Args()[0]
@@ -618,14 +706,12 @@ func (resyncPeriod *resync) changeWatcher() (err error) {
 	}
 
 	// Create a ticker to do a full node/pod labels pull
-	tickerPull := time.NewTicker(time.Second * time.Duration(resyncPeriod.now))
+	resyncPeriod := pullResyncPeriod()
+	tickerPull := time.NewTicker(time.Second * time.Duration(resyncPeriod))
 	defer tickerPull.Stop()
-	glog.Infof("Resync period to pull node/pod labels: %d [s]", resyncPeriod.now)
+	glog.Infof("Resync period to pull node/pod labels: %d [s]", resyncPeriod)
 
-	// When the first pod pull should happen; try to avoid lockstep
-	tuned.setNextPodLabelsPullTime(time.Now().Unix())
-
-	// Pull node and pod labels before entering the loop; node labels would otherwise be fetched after resyncPeriod.now
+	// Pull node and pod labels before entering the loop; node labels would be fetched after resyncPeriod
 	if err := pullLabels(clientset, &tuned, nodeName); err != nil {
 		return err
 	}
@@ -655,12 +741,57 @@ func (resyncPeriod *resync) changeWatcher() (err error) {
 		}
 	}
 
+	l, err := newUnixListener(openshiftTunedSocket)
+	if err != nil {
+		return fmt.Errorf("Cannot create %q listener: %v", openshiftTunedSocket, err)
+	}
+	defer func() {
+		lStop = true
+		l.Close()
+	}()
+
+	sockConns := make(chan sockAccepted, 1)
+	go func() {
+		for {
+			conn, err := l.Accept()
+			if lStop {
+				// The listener was closed on the return from mainLoop(); exit the goroutine
+				return
+			}
+			sockConns <- sockAccepted{conn, err}
+		}
+	}()
+
 	for {
 		select {
 		case <-done:
+			// Termination signal received, stop
 			glog.V(2).Infof("changeWatcher done")
+			if err := tunedStop(nil); err != nil {
+				glog.Errorf("%s", err.Error())
+			}
 			return nil
 
+		case s := <-sockConns:
+			if s.err != nil {
+				return fmt.Errorf("Connection accept error: %v", err)
+			}
+
+			buf := make([]byte, len("stop"))
+			nr, _ := s.conn.Read(buf)
+			data := buf[0:nr]
+
+			if string(data) == "stop" {
+				if err := tunedStop(&s); err != nil {
+					glog.Errorf("%s", err.Error())
+				}
+				return nil
+			}
+
+		case <-tunedExit:
+			cmd = nil // cmd.Start() cannot be used more than once
+			return fmt.Errorf("Tuned process exitted")
+
 		case fsEvent := <-wFs.Events:
 			glog.V(2).Infof("fsEvent")
 			// Ignore Write and Create events, wait for the removal of the old ConfigMap to trigger reload
@@ -684,13 +815,6 @@ func (resyncPeriod *resync) changeWatcher() (err error) {
 			if err := pullLabels(clientset, &tuned, nodeName); err != nil {
 				return err
 			}
-			if resyncPeriod.now/2 >= resyncPeriod.min {
-				// we've increased the original resyncPeriod due to errors; there was a successful
-				// pull now, converge back to the original resyncPeriod.min value
-				resyncPeriod.now /= 2
-				glog.V(1).Infof("Lowering resyncPeriod to %d", resyncPeriod.now)
-				tickerPull = time.NewTicker(time.Second * time.Duration(resyncPeriod.now))
-			}
 
 		case <-tickerReload.C:
 			glog.V(2).Infof("tickerReload.C")
@@ -701,12 +825,15 @@ func (resyncPeriod *resync) changeWatcher() (err error) {
 	}
 }
 
-func retryLoop() (err error) {
-	var resyncPeriod resync
-	resyncPeriod.now = pullResyncPeriodWithJitter()
-	resyncPeriod.min = resyncPeriod.now
+func retryLoop(f func() error) (err error) {
+	var errs int
+	const (
+		errsMax              = 5
+		errsMaxWithinSeconds = 120
+	)
+	errsTimeStart := time.Now().Unix()
 	for {
-		err = resyncPeriod.changeWatcher()
+		err = f()
 		if err == nil {
 			break
 		}
@@ -718,19 +845,16 @@ func retryLoop() (err error) {
 		}
 
 		glog.Errorf("%s", err.Error())
-		resyncPeriod.now *= 2
-		glog.V(1).Infof("Increasing resyncPeriod to %d", resyncPeriod.now)
-		if resyncPeriod.now > resyncPeriodNodeMax {
-			glog.Errorf("Increased resyncPeriod (%d) beyond the maximum (%d), terminating...", resyncPeriod.now, resyncPeriodNodeMax)
-			break
-		}
-
-		select {
-		case <-done:
-			return nil
-		case <-time.After(time.Second * time.Duration(resyncPeriod.now)):
-			continue
+		if errs++; errs >= errsMax {
+			now := time.Now().Unix()
+			if (now - errsTimeStart) <= errsMaxWithinSeconds {
+				glog.Errorf("Seen %d errors in %d seconds, terminating...", errs, now-errsTimeStart)
+				break
+			}
+			errs = 0
+			errsTimeStart = time.Now().Unix()
 		}
+		time.Sleep(time.Second * sleepRetry)
 	}
 	return err
 }
@@ -741,7 +865,7 @@ func main() {
 	logsCoexist()
 
 	if *boolVersion {
-		fmt.Fprintf(os.Stderr, "%s %s\n", PNAME, version)
+		fmt.Fprintf(os.Stderr, "%s %s\n", programName, version)
 		os.Exit(0)
 	}
 
@@ -750,8 +874,13 @@ func main() {
 		os.Exit(1)
 	}
 
+	err := openshiftTunedPidFileWrite()
+	if err != nil {
+		panic(err.Error())
+	}
+
 	sigs := signalHandler()
-	err := retryLoop()
+	err = retryLoop(changeWatcher)
 	signal.Stop(sigs)
 	if err != nil {
 		panic(err.Error())
